package filegateway

import (
	"context"
	"fmt"
	"io"
	"os"
	"path"
	"time"

	"github.com/pkg/sftp"
	"github.com/rs/zerolog"
	"golang.org/x/crypto/ssh"
	xbus "xbus.io/go-xbus/v4"
)

func sftpFilesystemFactory(options FilesystemOptions, log zerolog.Logger) (Filesystem, error) {
	if options.Host == "" {
		return nil, fmt.Errorf(
			"Invalid host for a sftp:// uri. Expects a non-mpty host")
	}
	if options.Port == "" {
		options.Port = "22"
	}
	fs := SFtpFilesystem{
		FilesystemOptions: options,
		address:           options.Host + ":" + options.Port,
		clientConfig:      ssh.ClientConfig{User: options.Username},
	}
	if options.HasPassword {
		fs.clientConfig.Auth = append(fs.clientConfig.Auth, ssh.Password(options.Password))
	}
	if pkey, err := options.GetCustomStringD("private-key", ""); err == nil {
		if pkey != "" {
			signer, err := ssh.ParsePrivateKey([]byte(pkey))
			if err != nil {
				return nil, fmt.Errorf("Error parsing 'private-key': %s", err)
			}
			fs.clientConfig.Auth = append(fs.clientConfig.Auth, ssh.PublicKeys(signer))
		}
	} else {
		return nil, err
	}
	if hostkey, err := options.GetCustomStringD("host-pubkey", ""); err == nil {
		if hostkey == "" {
			return nil, fmt.Errorf("'host-pubkey' is empty")
		}
		key, _, _, _, err := ssh.ParseAuthorizedKey([]byte(hostkey))
		if err != nil {
			return nil, fmt.Errorf("Error parsing 'host-pubkey': %s", err)
		}
		fs.clientConfig.HostKeyCallback = ssh.FixedHostKey(key)
	} else {
		return nil, err
	}
	if err := options.CustomStringD("basedir", &(fs.basedir), "."); err != nil {
		return nil, err
	}
	var (
		maxconn   int
		keepAlive time.Duration
	)
	if err := options.CustomIntD("maxconn", &maxconn, 2); err != nil {
		return nil, err
	}
	{
		var value string
		if err := options.CustomStringD("keep-alive", &value, "1m"); err != nil {
			return nil, err
		}
		duration, err := time.ParseDuration(value)
		if err != nil {
			return nil, err
		}
		keepAlive = duration
	}
	fs.client = NewSFtpPool(fs.address, &fs.clientConfig, 2, keepAlive)
	return &fs, nil
}

// SFtpFilesystem stores files on the local filesystem
type SFtpFilesystem struct {
	FilesystemOptions
	address      string
	clientConfig ssh.ClientConfig
	basedir      string

	client *SFtpPool
}

type sftpFileWriter struct {
	*sftp.File
	client *sftp.Client
	pool   *SFtpPool
}

// Close the sftp connection
func (fw sftpFileWriter) Close() error {
	err := fw.File.Close()
	fw.pool.Release(fw.client)
	return err
}

// Write ...
func (fs *SFtpFilesystem) Write(filepath string, size int) (io.WriteCloser, error) {
	client, err := fs.client.Get()
	if err != nil {
		return nil, err
	}
	filepath = path.Join(fs.basedir, filepath)
	dirname := path.Dir(filepath)
	if _, err := client.Stat(dirname); os.IsNotExist(err) {
		if err := client.MkdirAll(dirname); err != nil {
			fs.client.Release(client)
			return nil, err
		}
	}
	file, err := client.OpenFile(filepath, os.O_WRONLY|os.O_CREATE|os.O_TRUNC)
	if err != nil {
		fs.client.Release(client)
		return nil, err
	}
	return sftpFileWriter{file, client, fs.client}, nil
}

type sftpFileReader struct {
	*sftp.File
	client *sftp.Client
	pool   *SFtpPool
}

func (fr sftpFileReader) Close() error {
	err := fr.File.Close()
	fr.pool.Release(fr.client)
	return err
}

// Read returns an opened File
func (fs *SFtpFilesystem) Read(filepath string) (io.ReadCloser, error) {
	client, err := fs.client.Get()
	if err != nil {
		return nil, err
	}

	filepath = path.Join(fs.basedir, filepath)

	file, err := client.Open(filepath)
	if err != nil {
		fs.client.Release(client)
		return nil, err
	}

	return sftpFileReader{file, client, fs.client}, nil
}

// Rename renames a file
func (fs *SFtpFilesystem) Rename(oldpath, newpath string) error {
	oldpath = path.Join(fs.basedir, oldpath)
	newpath = path.Join(fs.basedir, newpath)
	return fs.client.Exec(func(client *sftp.Client) error {
		return client.Rename(oldpath, newpath)
	})
}

// Delete removes a file
func (fs *SFtpFilesystem) Delete(filepath string) error {
	filepath = path.Join(fs.basedir, filepath)
	return fs.client.Exec(func(client *sftp.Client) error {
		return client.Remove(filepath)
	})
}

// Observe naïvely do a file list every pollingInterval
func (fs *SFtpFilesystem) Observe(
	ctx context.Context, pattern string, paths chan<- string, logger xbus.Logger,
) error {
	log := zerolog.Ctx(ctx)
	pattern = path.Join(fs.basedir, pattern)
	fo := NewFileObserver(ctx, pattern, fs.basedir, paths)

	doCheck := func(first bool) error {
		matches, err := fs.client.Glob(pattern)
		if err != nil {
			log.Err(err).Msg("sftp: Error listing files")
			return err
		}
		if first {
			// first call to Glob succeeded, we can say we are ready
			select {
			case paths <- "":
			case <-ctx.Done():
				return ctx.Err()
			}
		}
		if err := fo.AddFiles(matches...); err != nil {
			log.Err(err).Msg("sftp: Error emitting files")
			return err
		}
		return nil
	}

	if err := doCheck(true); err != nil {
		return fmt.Errorf("sftp: first check failed: %w", err)
	}

	var errcount int
	for {
		var sleep time.Duration
		switch errcount {
		case 0:
			sleep = fs.PollingInterval
		case 1:
			sleep = 125 * time.Millisecond
		default:
			sleep *= 2
		}
		if sleep > fs.PollingInterval {
			sleep = fs.PollingInterval
		}
		if errcount == 1 {
			log.Error().Msgf("sftp: observe failed, will retry in %s", sleep)
		}
		if errcount > 1 {
			log.Error().Msgf("sftp: observe failed %d time, will retry in %s", errcount, sleep)
		}
		select {
		case <-time.After(sleep):
		case <-ctx.Done():
			return ctx.Err()
		}
		if err := doCheck(false); err != nil {
			errcount++
		} else {
			errcount = 0
		}
	}
}

// WaitForDeletion wath the given path every second until it does not exist
// anymore
func (fs *SFtpFilesystem) WaitForDeletion(ctx context.Context, filepath string) error {
	filepath = path.Join(fs.basedir, filepath)
	for {
		if _, err := fs.client.Stat(filepath); err != nil {
			if os.IsNotExist(err) {
				return nil
			}
			return err
		}
		select {
		case <-time.After(fs.PollingInterval):
		case <-ctx.Done():
			return ctx.Err()
		}
	}
}

func init() {
	MustRegisterFilesystem("sftp", sftpFilesystemFactory)
}
