package stdlib

import (
	"context"
	"errors"
	"io"

	"xbus.io/go-xbus/v4/envelope"
)

// ErrTooManyMessages is returned with the envelope contains too many messages
var ErrTooManyMessages = errors.New("too many messages")

// UniqueMessageContext returns the unique message of an envelope
func UniqueMessageContext(ctx context.Context, env envelope.Envelope) (envelope.Message, error) {
	defer runIfContextDone(ctx, func() {
		_ = env.CloseWithError(ctx.Err())
	})()

	msg, err := env.NextMessage(true)

	if err != nil {
		return nil, err
	}

	if _, err := env.NextMessage(true); err != io.EOF {
		return nil, ErrTooManyMessages
	}

	return msg, nil
}

// UniqueStringMessageContext returns the unique message of an envelope and its content as a string
func UniqueStringMessageContext(ctx context.Context, env envelope.Envelope) (envelope.Message, string, error) {
	msg, err := UniqueMessageContext(ctx, env)
	if err != nil {
		return nil, "", err
	}

	defer runIfContextDone(ctx, func() {
		_ = env.CloseWithError(ctx.Err())
	})()

	body, err := envelope.ReadAllAsString(msg)
	if err != nil {
		return nil, "", err
	}

	return msg, body, nil
}

func runIfContextDone(ctx context.Context, cb func()) func() {
	done := make(chan struct{})
	go func() {
		select {
		case <-ctx.Done():
			cb()
		case <-done:
		}
	}()
	return func() {
		close(done)
	}
}
