package routing

import (
	"context"
	"fmt"

	libjq_go "github.com/flant/libjq-go"
	"github.com/flant/libjq-go/pkg/jq"
	stdlib "orus.io/xbus/xbus-stdlib/lib"
	"orus.io/xbus/xbus-stdlib/service"

	"xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
)

// NewDeleter initialize a Deleter.
func NewDeleter(actor *api.Actor, settings service.Settings) *Deleter {
	return &Deleter{
		settings: settings,
	}
}

// NewDeleterAsWorker initialize a Deleter and returns it as ActorService.
func NewDeleterAsWorker(actor *api.Actor, settings service.Settings, log xbus.Logger) (service.Worker, error) {
	return NewDeleter(actor, settings), nil
}

// Deleter is a Worker that delete entries in the routing table.
type Deleter struct {
	settings service.Settings

	extractKeyJq *jq.JqProgram

	table TableStore
}

// Startup setup the routing table and compiles the jq programs.
func (u *Deleter) Startup() error {
	table, err := NewTableStore("store", u.settings)
	if err != nil {
		return err
	}

	extractKeyJq, err := u.settings.String("extract-key-jq")
	if err != nil {
		return fmt.Errorf("routing deleter: %w", err)
	}

	extractKeyJqProg, err := libjq_go.Jq().Program(extractKeyJq).Precompile()
	if err != nil {
		return fmt.Errorf("routing deleter: %w", err)
	}

	u.extractKeyJq = extractKeyJqProg
	u.table = table

	return nil
}

// Shutdown unsuscribes and close the DB connection.
func (u *Deleter) Shutdown() error {
	if err := u.table.Close(); err != nil {
		return fmt.Errorf("routing deleter shutdown: %w", err)
	}

	return nil
}

// Process ...
func (u *Deleter) Process(
	ctx context.Context, req *api.ActorProcessRequest,
	inputs service.Inputs, outputs service.Outputs, log xbus.Logger,
) error {
	env, err := inputs.Get(api.DefaultInput)
	if err != nil {
		return fmt.Errorf("routing deleter: %w", err)
	}

	_, body, err := stdlib.UniqueStringMessageContext(ctx, env)
	if err != nil {
		return fmt.Errorf("routing deleter: %w", err)
	}

	key, err := u.extractKeyJq.RunRaw(body)
	if err != nil {
		return fmt.Errorf("routing deleter: %w", err)
	}

	if err := u.table.Delete(key); err != nil {
		return fmt.Errorf("routing deleter: %w", err)
	}

	if err := outputs.Forward(api.DefaultOutput, env.ID()); err != nil {
		return fmt.Errorf("routing deleter: %w", err)
	}

	return nil
}

func init() {
	if err := service.RegisterWorker("routing.table-deleter", NewDeleterAsWorker); err != nil {
		panic(err)
	}
}
