package service

import (
	"context"
	"fmt"

	"xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
)

func catchErrors(log xbus.Logger) {
	if r := recover(); r != nil {
		log.Error().Interface("error", r).Stack().Msg("Caught error")
	}
}

// WorkerFunc is a function implementing a consumer
type WorkerFunc func(context.Context, *api.ActorProcessRequest, Inputs, Outputs, xbus.Logger) error

// Worker ...
type Worker interface {
	Process(context.Context, *api.ActorProcessRequest, Inputs, Outputs, xbus.Logger) error
}

// WorkerFactory is a factory of Worker
type WorkerFactory func(*api.Actor, xbus.Settings, xbus.Logger) (Worker, error)

// RegisterWorker registers a consumer
func RegisterWorker(name string, workerFactory WorkerFactory) (err error) {
	defer func() {
		if r := recover(); r != nil {
			if e, ok := r.(error); ok {
				err = fmt.Errorf("register worker: %w", e)
			} else {
				err = fmt.Errorf("register worker: %v", r) //nolint:goerr113
			}
		}
	}()
	xbus.RegisterActorService(name, newWorkerServiceFactory(workerFactory))
	return
}

func MustRegisterWorker(name string, workerFactory WorkerFactory) {
	if err := RegisterWorker(name, workerFactory); err != nil {
		panic(err)
	}
}

// RegisterWorkerFunc registers a consumer
func RegisterWorkerFunc(name string, process WorkerFunc) error {
	return RegisterWorker(name, func(*api.Actor, xbus.Settings, xbus.Logger) (Worker, error) { return process, nil })
}

// MustRegisterWorkerFunc registers a consumer
func MustRegisterWorkerFunc(name string, process WorkerFunc) {
	if err := RegisterWorkerFunc(name, process); err != nil {
		panic(err)
	}
}

// UnregisterWorker unregisters a worker
func UnregisterWorker(name string) {
	xbus.UnregisterActorService(name)
}

// Process calls the function
func (fn WorkerFunc) Process(ctx context.Context, req *api.ActorProcessRequest, inputs Inputs, outputs Outputs, log xbus.Logger) error {
	return fn(ctx, req, inputs, outputs, log)
}

func newWorkerServiceFactory(factory WorkerFactory) func(*xbus.Actor, xbus.Settings) (xbus.ActorService, error) {
	return func(actor *xbus.Actor, settings xbus.Settings) (xbus.ActorService, error) {
		apiActor := actor.Options.Actor
		worker, err := factory(apiActor, settings, actor.Log)
		if err != nil {
			return nil, err
		}
		return &actorService{
			actor:   actor,
			service: worker,
			process: workerProcess(worker),
		}, nil
	}
}

func workerProcess(worker Worker) func(
	ctx context.Context, svc *actorService, req *api.ActorProcessRequest,
	send func(*api.ActorProcessingState),
) error {
	return func(
		ctx context.Context, svc *actorService, req *api.ActorProcessRequest,
		send func(*api.ActorProcessingState),
	) error {
		inputs := realInputs{
			ctx:       ctx,
			client:    svc.actor.Client,
			actorID:   svc.actor.ID,
			actorKind: "worker",
			inputs:    req.Inputs,
		}
		outputs := realOutputs{
			ctx:               ctx,
			actor:             svc.actor,
			processingContext: req.Context,
		}
		return worker.Process(ctx, req, inputs, outputs, svc.actor.Log)
	}
}
